open Util

open Kernel

type t =
  | Skip
  | Tok of Pos.range * bool * Value.t * Value.t Compound.t
  | Eq of Pos.range * Value.t * Value.t
  | Call of Pos.range * Value.t Compound.t
  | Ask of Pos.range * Value.t * Var.t Compound.t * t
  | Exists of Pos.range * Var.t * t
  | Par of t * t

let rec big_par l =
  match l with
    | [] -> Skip
    | [a] -> a
    | h :: t -> Par (h, big_par t)

let rec big_exists p l a =
  match l with
    | [] -> a
    | h :: t -> Exists (p, h, big_exists p t a)

let rec big_exists_pos l a =
  match l with
    | [] -> a
    | (p, h) :: t -> Exists (p, h, big_exists_pos t a)

type point = node ref
and node =
  | Link of point
  | Desc of desc
and desc =
  | Extern of Var.t
  | Fresh of Pos.range * Var.t option
  | Term of point Term.t

let of_desc d = ref (Desc d)

let of_term t = of_desc (Term t)

let of_extern v = of_desc (Extern v)

let rec find v =
  match !v with
    | Link v' ->
        let (r, _) as result = find v' in
        if not (v' == r) then
          v := Link r;
        result
    | Desc d -> (v, d)

type env = {
  parent: Var.t -> point;
  mutable calls: (Pos.range * point Compound.t) list;
  mutable vars: (Pos.range * Var.t) list;
  mutable toks: (Pos.range * bool * point * point Compound.t) list;
  mutable asks:
    (point String.Map.t * Pos.range * point * Var.t Compound.t * t) list;
  mutable eqs: (Pos.range * point * point) list;
}

let rec to_val env v =
  let (r, d) = find v in
  match d with
    | Extern v -> Value.Var v
    | Fresh (p, o) ->
        let v =
          match o with
            | None ->
                let v = Var.fresh () in
                env.vars <- (p, v) :: env.vars;
                r := Desc (Fresh (p, Some v));
                v
            | Some v -> v in
        Value.Var v
    | Term t -> Value.Term (Term.map (to_val env) t)

let to_var env v =
  match to_val env v with
    | Value.Var v -> v
    | _ -> failwith "to_var"

let unify env pos v0 v1 =
  let (r0, d0) = find v0 and (r1, d1) = find v1 in
  if r0 == r1 then
    Warn.redundent_constraint pos
  else
    match d0, d1 with
      | Fresh _, _ -> r0 := Link r1
      | _, Fresh _ -> r1 := Link r0
      | Term (Term.Const c0), Term (Term.Const c1) ->
          if not (Const.equal c0 c1) then
            Err.constraint_always_inconsistent pos;
      | _ ->
          env.eqs <- (pos, r0, r1) :: env.eqs

let rec get_val parent ctx v =
  match v with
    | Value.Var v ->
        begin
          try String.Map.find v ctx
          with Not_found -> parent v
        end
    | Value.Term t -> of_term (Term.map (get_val parent ctx) t)

let rec split env ctx a =
  match a with
    | Skip -> ()
    | Tok (pos, to_consume, m, t) ->
        let m' = get_val env.parent ctx m in
        let t' = Compound.map (get_val env.parent ctx) t in
        env.toks <- (pos, to_consume, m', t') :: env.toks
    | Eq (pos, v0, v1) ->
        unify env pos (get_val env.parent ctx v0) (get_val env.parent ctx v1)
    | Call (pos, t) ->
        env.calls <- (pos, Compound.map (get_val env.parent ctx) t) :: env.calls
    | Ask (r, m, t, a0) ->
        let m' = get_val env.parent ctx m in
        env.asks <- (ctx, r, m', t, a0) :: env.asks
    | Exists (p, x, a0) ->
        let x' = ref (Desc (Fresh (p, None))) in
        split env (String.Map.add x x' ctx) a0
    | Par (a0, a1) ->
        split env ctx a0;
        split env ctx a1

let mk_call env (pos, t) = Agent.Call (pos, Compound.map (to_val env) t)

let mk_tok env (pos, to_consume, m, t) =
  Agent.Tell (pos, to_consume, to_var env m, Compound.map (to_val env) t)

let mk_eq env (pos, a, b) =
  let m = Files.use_module pos ["kernel"; "eq"] in
  let t = Compound.mk "make" [| to_val env a; to_val env b |] in
  Agent.Tell (pos, false, m, t)

let rec make_extern v =
  match v with
    | Value.Var v -> of_extern v
    | Value.Term t -> of_term (Term.map make_extern t)

let to_extern env v =
  make_extern (to_val env v)

let mk_ask env to_kernel (ctx, r, m, t, a) =
  let f (rev_new_args, ctx) v =
    let v' = Var.fresh () in
    (v' :: rev_new_args, String.Map.add v (of_extern v') ctx) in
  let (rev_new_args, ctx) = Array.fold_left f ([], ctx) t.Compound.args in
  let t' = Compound.mk (Compound.f t) (Array.of_list (List.rev rev_new_args)) in
  let new_parent v =
    try to_extern env (String.Map.find v ctx)
    with Not_found -> env.parent v in
  Agent.Ask (r, to_var env m, t', to_kernel new_parent a)

let rec to_kernel parent t =
  let env = {
    parent = parent;
    vars = [];
    calls = [];
    toks = [];
    asks = [];
    eqs = [];
  } in
  split env String.Map.empty t;
  let agent_calls = Agent.big_par (List.map (mk_call env) env.calls) in
  let agent_toks = Agent.big_par (List.map (mk_tok env) env.toks) in
  let agent_asks = Agent.big_par (List.map (mk_ask env to_kernel) env.asks) in
  let agent_eqs = Agent.big_par (List.map (mk_eq env) env.eqs) in
  Agent.big_exists (List.map snd env.vars)
    (Agent.big_par [agent_calls; agent_toks; agent_asks; agent_eqs])

let to_kernel t = to_kernel of_extern t
