open Util

open Interpret

open Data.Type

open Format

open Kernel

let show_version () =
  printf "\
SiLCC v%s
Thierry.Martinez@@inria.fr
Copyright (C) 2010 INRIA

Institut National de Recherche en Informatique et en Automatique,
Centre de recherche de Paris-Rocquencourt, France

SiLCC comes with ABSOLUTELY NO WARRANTY.
You may redistribute copies of SiLCC
under the terms of the GNU General Public License.
For more information about these matters, see the files named COPYING.
" Version.version

let show_help () =
  printf "\
Usage:
%s [OPTIONS ...] [FILE ...] [--arg argument ...]

Compile the given files and run them as LCC agents.
If no files are given, run the top-level.

Options:" Sys.argv.(0);
  Cmdline.show_help std_formatter;
  printf "\

Optimizations (comma-separated, may be prefixed by `<number of times>*'):
";
  Opt.show_help std_formatter;
  printf "\

Information options (should not be mixed with other arguments):
  -h, --help \tdisplay this list of options
  -v, --version \tprint SiLCC version
"

let execute = ref true

let compile_only () =
  execute := false

let () =
  Cmdline.register ~short:'c' ~long:"compile-only" (Cmdline.Zero compile_only)
    ~help:"compile (but do not execute) the agents"

let force = ref false

let force_execution () =
  force := true

let () =
  Cmdline.register ~short:'f' ~long:"force-execution"
    (Cmdline.Zero force_execution)
    ~help:"force the execution even in presence of warnings"

let print_time = ref false

let enable_print_time () =
  print_time := true

let () =
  Cmdline.register ~short:'t' ~long:"time"
    (Cmdline.Zero enable_print_time)
    ~help:"print compilation and execution times on standard error"

let print_statistics = ref false

let enable_print_statistics () =
  print_statistics := true

let () =
  Cmdline.register ~long:"statistics"
    (Cmdline.Zero enable_print_statistics)
    ~help:"print statistics on standard error after execution"

let main () =
  Builtins.dummy ();
  Expat_builtins.dummy ();
  Llvm_builtins.dummy ();
  match Sys.argv with
    | [| _; ("-v" | "--version") |] ->
        show_version ()
    | [| _; ("-h" | "--help") |] ->
        show_help ()
    | _ ->
        Cmdline.parse (fun filename -> Files.add filename);
        if Files.is_empty () then
          Top.run ();
        let compile_start_time = Sys.time () in
        let opt = Compile.all () in
        let (am, global_vars) = Am.of_opt opt Am.indexed_empty in
        if !print_time then
          Format.fprintf Format.err_formatter "Compilation time: %f@."
            ((Sys.time ()) -. compile_start_time);
        if !execute then
          begin
            if !Warn.has_warning && not !force then
              begin
                Format.fprintf Format.err_formatter "
Execution disabled in presence of warning.
Use the `-f' option to force execution.@.";
                exit 1
              end
            else
              begin
                let execution_start_time = Sys.time () in
                let global_array =
                  Array.make global_vars.Am.count Interpret.dummy in
                let _ = Interpret.execute am global_array in
                if !print_time then
                  Format.fprintf Format.err_formatter "@.Execution time: %f@."
                    ((Sys.time ()) -. execution_start_time);
                if !print_statistics then
                  Interpret.print_statistics Format.err_formatter
              end
          end

let () =
  if not !Sys.interactive then
    try main ()
    with Err.E (pos, msg) ->
      Err.print pos msg;
      exit 1
